{-# OPTIONS -fglasgow-exts #-}

-- Dynamically scoped variable in Haskell
-- We're using the delimited continuation framework by 
-- Dybvig, Peyton-Jones, and Sabry (CC monad transformers).
-- See caml-dynvar.ml for comparison

module Dynvar where

import CC_FrameT
import Control.Monad.Trans
import Control.Monad.Identity


-- Delimited control operators that we use here

abortP :: Monad m => Prompt r a -> CC r m a -> CC r m b
abortP p e = letSubCont p (\_ -> e)

promptP :: Monad m => (Prompt r a -> CC r m a) -> CC r m a
promptP f = do p <- newPrompt; pushPrompt p (f p)

shiftP :: Monad m => 
	  Prompt r b -> ((CC r m a -> CC r m b) -> CC r m b) -> CC r m a
shiftP p f = letSubCont p $ \sk -> 
               pushPrompt p (f (\c -> 
                 pushPrompt p (pushSubCont sk c)))


-- the type of dynamic variables
newtype Dynvar r m a = Dynvar (Prompt r (a->CC r m a))

dnew :: Monad m => CC r m (Dynvar r m a)
dnew = newPrompt >>= return . Dynvar

dref :: Monad m => Dynvar r m a -> CC r m a
dref (Dynvar p) = 
    shiftP p (\f -> return $ \v -> f (return v) >>= ( $ v))

dset :: Monad m => Dynvar r m a -> a -> CC r m a
dset (Dynvar p) newv = 
    shiftP p (\f -> return $ \v -> f (return v) >>= ( $ newv))

-- dupp p V corresponds to $V p$ in the paper
dupp :: Monad m => Dynvar r m a -> (a -> CC r m b) -> CC r m b
dupp (Dynvar p) g = 
    shiftP p (\f -> return $ \y -> g y >>= f . return >>= ( $ y))

dlet :: Monad m => Dynvar r m a -> a -> CC r m b -> CC r m b
dlet (Dynvar p) v body =
    do
    q <- newPrompt
    pushPrompt q (
       pushPrompt p (body >>= (\z -> abortP q (return z)))
       >>= ( $ v) >>= undefined)


-- Example 1 from the paper

testc = runIdentity $ runCC (
   do
   p <- dnew
   dlet p 0 (do
	     let f = dref p
	     x <- f
	     y <- dlet p 1 f
	     z <- f
	     return (x,y,z)))

-- (0,1,0)


-- Extended example with mutation
test12 = runIdentity $ runCC (
   do
   p <- dnew
   dlet p 1 (do
	     v1 <- dref p
	     v2 <- dlet p 2 (do
			     v3 <- dset p 12
			     v4 <- dref p
			     return (v3,v4))
	     v5 <- dref p
	     return (v1,v2,v5)))
-- (1,(2,12),1)



-- Run the motivating examples from Section 4 of the paper and show
-- that in the present implementation, we obtain the expected results.

test_eq4 = runIdentity $ runCC (
   do
   p  <- dnew
   p0 <- newPrompt
   dlet p 1 (pushPrompt p0 (dref p)))
--  ==> 1

test_eq5 = runIdentity $ runCC (
   do
   p  <- dnew
   p0 <- newPrompt
   dlet p 1 (
     pushPrompt p0 (
      dlet p 2 (
	shiftP p0 (\f -> dref p)))))
--  ==> 1

data RK r m a b = J (CC r m a -> CC r m (RK r m a b)) | R b

test_eq6 = runIdentity $ runCC (
   do
   p  <- dnew
   q  <- dnew
   p0 <- newPrompt
   (\ (J f) -> dlet p 2 $ dlet q 20 (f (return 0) >>= \ (R x) -> return x)) =<<
    (dlet p 1 $ pushPrompt p0 $ dlet q 10 
			      ((shiftP p0 (return . J)) >>=
			       (\x -> liftM2 (+) (dref p) (dref q)) >>=
			       return . R)))

--  ==> 12


-- Now check our old `counter-example', which exhibited the problem
-- (in the Scheme and SML code)


tf dv title = 
    do
    liftIO $ putStrLn title
    dref dv >>= liftIO . print
    dref dv >>= liftIO . print

test2 :: IO () = runCC (
   do
   p0  <- newPrompt
   dv0 <- dnew
   liftIO $ putStrLn "test2"
   dlet dv0 1 $
     do
     J f <- pushPrompt p0 $
	      dlet dv0 2 $
		do
		f1 <- shiftP p0 (\f -> tf dv0 "inside shift" >> (return $ J f))
		f1 "inside again"
		return $ R ()
     tf dv0 "outside"
     f (return (tf dv0))
     tf dv0 "outside again")

{-
test2
inside shift
1
1
outside
1
1
inside again
2
2
outside again
1
1
-}

test3 :: IO () = runCC (
   do
   p0  <- newPrompt
   dv0 <- dnew
   liftIO $ putStrLn "test3"
   dlet dv0 1 $
     do
     J f <-  dlet dv0 2 $
              pushPrompt p0 $
		do
		f1 <- shiftP p0 (\f -> tf dv0 "inside shift" >> (return $ J f))
		f1 "inside again"
		return $ R ()
     tf dv0 "outside"
     f (return (tf dv0))
     tf dv0 "outside again")


{-
test3
inside shift
2
2
outside
1
1
inside again
1
1
outside again
1
1
-}


-- ------------------------------------------------------------------------
-- Another extension: folding over the context *)

-- Ken's examples

dnil p body    = dlet p Nothing body
dcons p v body = dlet p (Just v) body
dmemberp p v = dupp p (maybe (return False)
		             (\y -> if v == y then return True
			               else dmemberp p v))

testa = runIdentity $ runCC (
   do
   p  <- dnew
   dnil p $ dcons p 1 $ dcons p 2 (
	liftM3 (,,) (dmemberp p 2) (dmemberp p 1) (dmemberp p 3)))

-- ==> (True,True,False)


{-
   Ken's nub example, first version (from the earlier version of the paper)
   nub is a function that removes duplicates from the input list,
   while maintaining the order otherwise: if an element occurs
   several times in the input list, only its _first_ occurrence remains
   in the output list.
   Our function below only assumes equality on the elements of the list;
   we do _not_ assume any order relation for the list elements and
   so can't use sort.
-}

nub lst = do
  p <- dnew
  let nub' [] = return []
      nub' (h:t) = do
		   seen_before <- dmemberp p h
		   if seen_before then nub' t
		      else dcons p h (liftM (h:) $ nub' t)
  dnil p $ nub' lst


test_nub = runIdentity $ runCC (
	      nub [1,1,1,1,1,1,1,1,1,1,3,2,1,1,1,1,1,1,2,1,1,1])

-- ==> [1,3,2]
