(** Dynamically scoped variable in Ocaml 
    We're using the CC (aka caml-shift) framework of multi-prompt
    delimited-continuation operators.

  $Id: caml-dynvar.ml,v 1.8 2006/04/07 00:42:53 oleg Exp $
*)

open Delimcc

(* The signature of the dynamically-scoped facility *)


(* First we define multi-prompt shift and abort. The latter is a convenient
   abbreviation 
   The operators below are _not_ trampolined ones!
*)

let shift p f = take_subcont p (fun sk () ->
  push_prompt p (fun () -> (f (fun v -> 
    push_prompt p (fun () -> push_subcont sk (fun () -> v))))))
;;
let abort p v = take_subcont p (fun _ () -> v)
;;


(* The first attempt, not very satisfactory *)

let dref p = shift p (fun x -> fun y -> x y y)
;;
let dlet p v body = (push_prompt p (fun () -> ((fun z y -> z) (body ())))) v
;;


let test1 = 
  let p = new_prompt () in
  dlet p 1 (fun () -> [dref p; dref p])
;;


(* The following is already a problem
let test11 = 
  let p = new_prompt () in
  dlet p 1 (fun () -> [dref p; dref p; dlet p 2 (fun () -> dref p)])
;;
let testc = 
  let p = new_prompt () in
   dlet p 0 (fun () -> 
   let x = f () in
   let y = dlet p 1 (fun () -> f ()) in
   let z = f () in
   (x,y,z))
;;
*)

module type Dynvar = sig
  type 'a dynvar
  val dnew : unit -> 'a dynvar
  val dref : 'a dynvar -> 'a
  val dset : 'a dynvar -> 'a -> 'a
  val dupp : 'a dynvar -> ('a -> 'w) -> 'w  (* dupp p V === V p in the paper*)
  val dlet : 'a dynvar -> 'a -> (unit -> 'w) -> 'w
end
;;


(* The function ign, whose application to a value never yields anything and
   never returns. That function is generally fun x -> omega
 
Here's one possible implementation
let rec fix f n = f (fix f) n;;

let ign x = fix (fun self y -> self y) x

And here's another:
*)
let ign x = failwith "cannot happen"
;;



(* A better attempt *)
let dlet p v body = 
  let q = new_prompt () in
  push_prompt q (fun () ->
    ign 
    ((push_prompt p (fun () -> ((fun z -> shift q (fun _ -> z)) (body ())))) v)
  )
;;


let test1 = 
  let p = new_prompt () in
  dlet p 1 (fun () -> [dref p; dref p])
;;

let test11 = 
  let p = new_prompt () in
  dlet p 1 (fun () -> [dref p; dref p; dlet p 2 (fun () -> dref p)])
;;

module DynvarI : Dynvar = struct
  type 'a dynvar = ('a -> 'a) Delimcc.prompt
  let dnew () = new_prompt ()
  let dref p = shift p (fun x -> fun v -> x v v)
  let dset p newv = shift p (fun x -> fun v -> x v newv)
  let dupp p g = shift p (fun f -> fun y -> f (g y) y)
  let dlet p v body = 
    let q = new_prompt () in
    push_prompt q (fun () ->
      ign 
       ((push_prompt p (fun () -> ((fun z -> shift q (fun _ -> z)) (body ()))))
	  v))
end
;;

open DynvarI
;;
let test12 =
  let p = dnew () in
  dlet p 1 (fun () ->
    let v1 = dref p in
    let v2 = dlet p 2 (fun () ->
      let v3 = dset p 12 in
      let v4 = dref p in
      (v3,v4)) in
    let v5 = dref p in
    (v1,v2,v5))
;;

(* (1, (2, 12), 1) *)


let testc = 
  let p = dnew () in
   dlet p 0 (fun () -> 
   let f = fun () -> dref p in
   let x = f () in
   let y = dlet p 1 (fun () -> f ()) in
   let z = f () in
   (x,y,z))
;;



(** Now check our old `counter-example', which exhibited the problem
    (in the Scheme and ML code)
*)


(*------------------------------------------------------------------------
 * For simplicity and consistency with other tests, we use only one
 * dynvar, so we don't bother with naming it explicitly. For
 * compatibility with the SML code (see *.sml files in this directory)
 * we don't make dlet a macro -- hence dlet takes a thunk as the second
 * argument.
 *)

(* We capture a recursive continuation. We could have avoided that,
   but we'd like the close correspondence with the other versions
   of our example.
*)

type ('a,'b) reck = 
    J of ('a -> ('a,'b) reck)
  | R of 'b


let dv0 = dnew ()

let tf title = 
  print_endline title;
  print_int (dref dv0); print_endline "";
  print_int (dref dv0); print_endline ""
;;

let test2 = 
  let p0 = new_prompt () in   (* prompt for the local control *)
  let () = print_endline "test2" in
  dlet dv0 1 (fun () ->
    let J f = push_prompt p0 (fun () ->
      dlet dv0 2 (fun () ->
	let f1 = shift p0 (fun f -> tf "inside shift"; J f) in
	f1 "inside again"; R ())) in
    let () = tf "outside" in
    let _ =  f tf in
    tf "outside again")
;;

(*
test2
inside shift
1
1
outside
1
1
inside again
2
2
outside again
1
1
*)

let test3 = 
  let p0 = new_prompt () in   (* prompt for the local control *)
  let () = print_endline "test3" in
  dlet dv0 1 (fun () ->
    let J f = 
      dlet dv0 2 (fun () ->
	push_prompt p0 (fun () ->
	let f1 = shift p0 (fun f -> tf "inside shift"; J f) in
	f1 "inside again"; R ())) in
    let () = tf "outside" in
    let _ =  f tf in
    tf "outside again")
;;

(*
test3
inside shift
2
2
outside
1
1
inside again
1
1
outside again
1
1
*)

(* Tests from the paper, Section 4 *)

let test_eq4 = (* eq. (4) *)
  let p = dnew () and p0 = new_prompt () in
  dlet p 1 (fun () -> push_prompt p0 (fun () -> dref p))
;;
(* ==> 1 *)

let test_eq5 = (* eq. (5) *)
  let p = dnew () and p0 = new_prompt () in
  dlet p 1 (fun () -> 
    push_prompt p0 (fun () ->
      dlet p 2 (fun () -> 
	shift p0 (fun f -> dref p))))
;;
(* ==> 1 *)

let test_eq6 = (* eq. (6) *)
  let p = dnew () and q = dnew () and p0 = new_prompt () in
  (fun (J f) -> 
    dlet p 2 (fun () -> 
      dlet q 20 (fun () -> match f 0 with R x -> x)))
    (dlet p 1 (fun () -> 
      push_prompt p0 (fun () ->
	dlet q 10 (fun () -> 
	  R ((fun x -> dref p + dref q) (shift p0 (fun f -> J f)))))))
;;
(* ==> 12 *)
    



(*------------------------------------------------------------------------*)
(* Another extension: folding over the context *)

(* Ken's examples *)

let dnil  p body   = dlet p None body
let dcons p v body = dlet p (Some v) body
let rec dmemberp p v = dupp p 
    (function 
      |	None -> false
      | Some y -> v == y || dmemberp p v)
;;

let testa =
  let p = dnew () in
  dnil p (fun () ->
    dcons p 1 (fun () ->
      dcons p 2 (fun () ->
	(dmemberp p 2, dmemberp p 1, dmemberp p 3))))
;;
(* (true,true,false) *)


(* Ken's nub example, first version (from the earlier version of the paper) *)
(* nub is a function that removes duplicates from the input list,
   while maintaining the order otherwise: if an element occurs
   several times in the input list, only its _first_ occurrence remains
   in the output list.
   Our function below only assumes equality on the elements of the list;
   we do _not_ assume any order relation for the list elements and
   so can't use sort.
*)
let nub lst =
  let p = dnew () in
  let rec nub' = 
    function 
      |	[] -> []
      |	(h::t) -> 
	  if dmemberp p h (* h has been seen before *)
	      then nub' t
	      else dcons p h (fun () -> (h :: nub' t))
  in dnil p (fun () -> nub' lst)
;;

let test_nub = nub [1;1;1;1;1;1;1;1;1;1;3;2;1;1;1;1;1;1;2;1;1;1]
;;
(* [1;3;2] *)


(* Ken's nub example *)
(* It also emulates the list as above -- but in a different, dual way,
   using a fold-like combinator.
   To be precise, our data structure is an abstract list, a pair
   of the type ('a -> 'a abstract_list -> 'a abstract_list,
                'a abstract_list)
   The first component is like 'cons' and the latter is like 'nil'
*)

let dfnil p body = 
   let apply_nil abstract_list = abstract_list in
   apply_nil (dlet p apply_nil body)
let dfcons p v body = 
   let apply_cons (cons,nil) = (cons,cons v nil) in
   apply_cons (dlet p apply_cons body)
;;

let nub2 lst =
  let p = dnew () in
   (* we need two realizations of the abstract list *)
  let regular_list = ((fun h t -> h::t),[]) in
  let rec dfmemberp v = (* this uses another realization of the list:
			   as a boolean *)
   let rec cons a nil = if a == v then true else dfmemberp v
   and nil = false in
   snd (dupp p (fun f -> f (cons,nil))) in
  let rec nub' = 
    function 
      |	[] -> regular_list
      |	(h::t) -> 
	  if dfmemberp h (* seen before *)
	      then nub' t
	      else dfcons p h (fun () -> nub' t)
  in snd (dfnil p (fun () -> nub' lst))
;;
let test_nub2 = nub2 [1;1;1;1;1;1;1;1;1;1;3;2;1;1;1;1;1;1;2;1;1;1]
;;
(* [1;3;2] *)
