; 		Control operators for delimited continuations
;
;  An implementation of the delimited continuation framework of Dybvig, Sabry,
;  Peyton-Jones for (R5RS + records) Scheme
;
;  This code is based on CC_Ref.hs -- the template for implementing delimited
;  continuations on a system permitting no inspection of the stack. We 
;  cannot check the stack for prompt marks.
; 
; This code extends Dybvig et al API with an extra function:
;   prompt-set? :: prompt -> Bool
; which tells if a particular prompt is set. We need this function to
; fully implement Chez Scheme parameters -- to be precise, the
; `default value' of the parameter (to be used when it is not dynamically set).
;
; This code is based on delimcc.ml -- the native implementation of
; delimited continuations for byte-code OCaml.
; We take an advantage of Scheme eq? and the lack of typing in Scheme
; and so we can dispense with marks and merge holes and dwa actions.
;
; $Id: cc-ref.scm,v 1.2 2006/03/09 08:46:17 oleg Exp $


; We use Petite Chez records; Other record systems are similar

; ------------------------------- Types

; Dynamic wind action + hole in the context
;      A component of the parallel stack, associated with each hole.
;  The list of DWAs `follows' the context and describes all active
;  holes in the context

(define-record dwa
  ((immutable set)     ; action
   (immutable reset)   ; action
   (immutable k)       ; the stack associated with dwa
  ))

;  The parallel stack 
; type pstack = dwa list ref

(define-record prompt () (((mutable holes) '())))

; The context between two holes.
;    It includes two stack pointers and the
;    corresponding copy of the parallel stack. The latter is a list
;    of DWAs in inverse order (the farthest DWA is at the bottom of the
;    chain).

(define-record subcont
  ((immutable k)
   (immutable chain)))

;------------------------------- The global State

(define ptop '())		; The top of the parallel stack


; ------------------------------- Utilities

; Pass the value to a continuation in the top-level context.
; This is a trampoline: we first reset the stack and then perform
; the application

(define go-apply
  ((call-with-current-continuation
     (lambda (k0)
       (lambda ()
	 (lambda (k v)
	   (k0 (lambda () (k v)))))))))

;  Operations on pstack. The invariant: `pop' and `head' operations are
;    never attempted on the empty stack.

(define (pstack-head) (car ptop))
(define (pstack-pop)
  (set! ptop (cdr ptop)))
(define (pstack-push dwa)
  (set! ptop (cons dwa ptop)))

;   Operations on prompts, which are stacks of holes.
;    Here, stack underflow is possible if the user forgot to set the prompt

(define (check-prompt p)
  (if (null? (prompt-holes p))
    (error 'check-prompt "No prompt set")))
(define (prompt-head p)
  (check-prompt p) (car (prompt-holes p)))
(define (prompt-pop p)
  (check-prompt p) (set-prompt-holes! p (cdr (prompt-holes p))))
(define (prompt-push p dwa)
  (set-prompt-holes! p (cons dwa (prompt-holes p))))


;  Plug the value `a' into the hole. Undo the hole's action
;  INV: ptop points to the DWA for the hole
(define (undo1 a)
  (let*
    ((dwa (pstack-head))
     (k (dwa-k dwa))
     (_ ((dwa-reset dwa))))  ; changes the stack at ptop
    (go-apply k a)))         ; does not return


; Unwind ptop until the specified hole (dwa, actually)
; leaving the latter on the stack. Return the list of removed
; dwas, in inverse order (that is, the most recently removed is first).
; As we remove a dwa, we execute its reset action

(define (unwind dwa acc)
  (let ((cdwa (pstack-head)))
    (if (eq? dwa cdwa) acc
      (begin
	((dwa-reset cdwa)) ; changes ptop
	(unwind dwa (cons cdwa acc))))))


; ------------------------------- Exported operations

(define new-prompt make-prompt)
(define (prompt-set? p) (not (null? (prompt-holes p))))

(define (push-prompt* p thunk)
 ((call-with-current-continuation  ; yields thunk....
    (lambda (k)
      (letrec ((dwa
		 (make-dwa
	           ; dwa-set
		   (lambda () (prompt-push p dwa) (pstack-push dwa))
	           ; dwa-reset
		   (lambda () (prompt-pop p) (pstack-pop))
		   k)))
	((dwa-set dwa))
	(let ((res (thunk)))
	  (undo1 (lambda () res))))))))


; letSubCont :: Prompt r b -> (SubCont r a b -> CC r b) -> CC r a
(define (take-subcont* p consumer)
  ((call-with-current-continuation
     (lambda (k)
       (let* 
	 ((hole (prompt-head p))       ; dwa for the prompt
	  (chain (unwind hole '())))   ; unwind till marked
         ; The prompt's dwa must now be at ptop
	 (undo1 (consumer (make-subcont k chain))))))))

; push sk and then evaluate thunk
(define (push-subcont* sk thunk)
  ((call-with-current-continuation
     (lambda (k)
      (letrec ((dwa
		 (make-dwa
	           ; dwa-set
		   (lambda () (pstack-push dwa))
	           ; dwa-reset
		   (lambda () (pstack-pop))
		   k)))
	((dwa-set dwa))
	(for-each (lambda (dwa) ((dwa-set dwa))) (subcont-chain sk))
	(go-apply (subcont-k sk) thunk))))))


; ------------------------------- Syntactic sugar

(define-syntax push-prompt
  (syntax-rules ()
    ((_ p e1 e2 ...) (push-prompt* p (lambda () e1 e2 ...)))))

(define-syntax abortP
  (syntax-rules ()
    ((_ p e) (take-subcont* p (lambda (dummy) (lambda () e))))))


(define-syntax take-subcont
  (syntax-rules ()
    ((_ p sk e1 e2 ...)
      (take-subcont* p (lambda (sk) (lambda () e1 e2 ...))))))

(define-syntax push-subcont
  (syntax-rules ()
    ((_ sk e1 e2 ...)
      (push-subcont* sk (lambda () e1 e2 ...)))))

; The reified continuation takes a value rather than an action
(define-syntax shiftV
  (syntax-rules ()
    ((_ p f e1 e2 ...)
      (take-subcont p sk
	(let ((f (lambda (v) (push-prompt p (push-subcont sk v)))))
	  (push-prompt p e1 e2 ...))))))

(define-syntax controlV
  (syntax-rules ()
    ((_ p f e1 e2 ...)
      (take-subcont p sk
	(let ((f (lambda (v) (push-subcont sk v))))
	  (push-prompt p e1 e2 ...))))))

