;	         A counter-example
; to show that an implementation of dynamic variables in scheme,
; SRFI-39, leads to the unexpected behavior in the presence of
; delimited continuations (shift/reset)

; This is R5RS code, tested with Petite Chez Scheme, SCM, and Scheme48.


;------------------------------------------------------------------------
; First, some preliminaries: SRFI-39
;
; The following code is taken verbatim from SRFI-39
; `Parameter objects'
; http://srfi.schemers.org/srfi-39/srfi-39.html
; The only addition is the debugging printing for dynamic-wind.

(define make-parameter
  (lambda (init . conv)
    (let ((converter
	    (if (null? conv) (lambda (x) x) (car conv))))
      (let ((global-cell
	      (cons #f (converter init))))
	(letrec ((parameter
		   (lambda new-val
		     (let ((cell (dynamic-lookup parameter global-cell)))
		       (cond ((null? new-val)
			       (cdr cell))
			 ((null? (cdr new-val))
			   (set-cdr! cell (converter (car new-val))))
			 (else ; this case is needed for parameterize
			   (converter (car new-val))))))))
	  (set-car! global-cell parameter)
	  parameter)))))

(define-syntax parameterize
  (syntax-rules ()
    ((parameterize ((expr1 expr2) ...) body ...)
      (dynamic-bind (list expr1 ...)
	(list expr2 ...)
	(lambda () body ...)))))

(define dynamic-bind
  (lambda (parameters values body)
    (let* ((old-local
	     (dynamic-env-local-get))
	    (new-cells
	      (map (lambda (parameter value)
		     (cons parameter (parameter value #f)))
		parameters
		values))
	    (new-local
	      (append new-cells old-local)))
      (dynamic-wind
	(lambda () (display "   dynamic-bind: setting to ")
	           (display new-local) (newline)
	  (dynamic-env-local-set! new-local))
	body
	(lambda () (display "   dynamic-bind: resetting to ")
	           (display old-local) (newline)
	  (dynamic-env-local-set! old-local))))))

(define dynamic-lookup
  (lambda (parameter global-cell)
    (or (assq parameter (dynamic-env-local-get))
                global-cell)))

(define dynamic-env-local '())

(define dynamic-env-local-get
  (lambda () dynamic-env-local))

(define dynamic-env-local-set!
  (lambda (new-env) (set! dynamic-env-local new-env)))


;------------------------------------------------------------------------
; The second part of preliminaries:
; Danvy/Filinski implementation of shift/reset, the Christmas present
; to Scheme48 authors.
(load "shift-reset.scm")


;------------------------------------------------------------------------
; For simplicity and consistency with other tests, we use only one
; dynvar, so we don't bother with naming it explicitly. For
; compatibility with the SML code (see *.sml files in this directory)
; we don't make dlet a macro -- hence dlet takes a thunk as the second
; argument.


; dereference the (implicit) dynamic variable
; Just invoke (dref), which will be bound via parameterize
; (define (dref) ...)
(define dref (make-parameter #f))


; Evaluate THUNK in the dynamic environment extended with the binding of the
; (implicit) dynvar to V
(define (dlet v thunk)
  (parameterize ((dref v)) (thunk)))

; The first test
(define (test1)
  (dlet 1
    (lambda ()
      (list (dref) (dref)
      (dlet 2 
	(lambda ()
	  (list (dref) (dref))))))))

(display (test1)) (newline)
; (1 1 (2 2))


; The counter-example itself. 
; Its code is the same as in the other dynvar counter-examples

; A sample function to print the current dynamic environment, twice
; (just for the heck of it).
(define (tf title)
  (display title) (newline)
  (display (dref)) (newline)
  (display (dref)) (newline))

; The first test: we capture a delimited continuation that 
; DOES include `dlet 2'
(define (test2)
  (newline) (display "test2") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
	      (reset
		(dlet 2 
		  (lambda ()
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test2)

; The second test: we now capture a delimited continuation that does NOT
; include `dlet 2'
(define (test3)
  (newline) (display "test3") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
		(dlet 2 
		  (lambda ()
	      (reset
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test3)

; We observe, test2 and test3 print the same results. But that is wrong!
; In test3, the dynamic environment in the case "inside again"
; (that is, when the captured continuation is reinstated)
; must be 1 rather than 2 because `dlet 2' is NOT included within the
; captured continuation.



;------------------------------------------------------------------------
;   Dynamic variables and threads: we demonstrate that if we use
; shift/reset to implement `threads', each thread has its own dynamic
; environment. This is similar to processes in Unix: address space
; of a process is private to the process and is not shared with other
; processes and the parent (from which it was `inherited').
; That seems like a good thing. However: shift/reset can be used for
; many other things rather than the implementation of threads.
; For those things, such an isolation of dynamic environment is inappropriate
; as it defeats the purpose of the dynamic environment: parameterizing
; the code (cf, `hostname', `current port', etc parameters).
; Moreau and other people argue that the reason dynvars are useful is because
; they are not captured in a closure. The behavior exhibited in test4
; below shows that the dynamic environment is essentially captured
; in thread's `closure'. Because the standard, dynamic-wind based
; implementation of dynvars `captures' all of the dynamic environment,
; it becomes very difficult to let some dynvars `float out'. Even if
; some dynvars should be private to a thread, some other dynvars should not.
; It is the programmer who should specify what should be private to a thread,
; and what should be allowed to `propagate' so that the code (including the
; code invoked by the current thread) can be parameterized. The dynamic-wind
; implementation makes such determination very difficult.
; There is another problem with the implementation: as we can observe 
; from the trace of dynamic-wind invocations, each `thread switching' 
; causes undoing of the dynlet along one continuation -- all the way 
; up to the thread staring point -- and redoing of all dynlet along 
; the other continuation. If the threads since forking have done a few dynlets,
; then _each_ thread switch will include undoing and redoing them all
; (even the ones whose effect is shadowed by the later dynlets).
; That means the thread switch is no longer light-weight!
; The effect is exacerbated if shift/reset is implemented so to prevent
; the memory leak (aka JAR hack): in that case, all active dynlets have to
; be undone, and all dynlets to be active in the thread re-done.


(define (test4)
  (newline) (display "test4: two threads, switching") (newline)
  (dlet 1
    (lambda ()
      (let*
	((thread1
	   (reset
	     (begin
	       (tf "thread1: starting")
	       (dlet 10
		 (lambda ()
		   (shift f f)		; suspend
		   (tf "thread1: continuing...")
		   (shift f f)		; suspend
		   (tf "thread1: finishing, after dlet..."))))))
	  (thread2
	    (dlet 2
	      (lambda ()
		(reset
		  (begin
		    (tf "thread2: starting")
		    (shift f f)		; suspend
		    (tf "thread2: continuing...")
		    (dlet 22
		      (lambda ()
			(shift f f)		; suspend
			(tf "thread2: finishing, after dlet..."))))))))
	  )
	(dlet 100
	  (lambda ()
	    (tf "main thread")
	    (let*
	      ((thread1 (thread1 #f))
	       (_ (tf "main thread"))
	       (thread2 (thread2 #f))
		)
	      (thread1 #f)
	      (thread2 #f)
	      )))))))
(test4)
; thread1: starting
; 1
; 1
; thread2: starting
; 2
; 2
; main thread
; 100
; 100
; thread1: continuing...
; 10
; 10
; main thread
; 100
; 100
; thread2: continuing...
; 2
; 2
; thread1: finishing, after dlet...
; 10
; 10
; thread2: finishing, after dlet...
; 22
; 22

