; This file presents a realistic example of how easy it is
; to encounter the pitfall of the common implementations
; of delimited continuations and dynamically-scoped variables.
; The two features, as commonly implemented, interact in unexpected ways.
;
; The features in question are exceptions and delimited continuations,
; both part of Scheme48, version 1.3.
;
; As was pointed out by Moreau, an exception handler can be understood as
; a dynvar. In Scheme48, this is literally the case: the file
;	scheme48-1.3/scheme/rts/exception.scm
; of Scheme48 distribution says:
;   Raising and handling conditions.
;   (fluid $exception-handlers) is a cell containing a list of 
;   handler procedures.
;
; To run the example, start scheme48 and do the following:
;
;   ,open srfi-34
;   ,open signals escapes
;   ,load =scheme48/misc/shift-reset.scm
;   ,load exceptions-shift.scm
;
; The first line loads the exception-handling interface,
; SRFI-34. The next two lines make available the control operators
; shift and reset.  The last line loads the present code.

; Our particular example uses delimited continuations to `invert' an
; enumerator over a collection that uses i/o (e.g., a database or a
; file). Doing I/O can always raise an exception. 
; We use exception handling to make sure that opened files are closed,
; whether our processing finished normally or abnormally.
; To avoid reading and writing files in this example, we simulate file
; i/o and use explicit 'raise' to raise an i/o `exception' at some moment.
;
;
; $Id: exceptions-shift.scm,v 1.1 2006/04/02 07:13:24 oleg Exp $


; Suppose we have a library (called LIB1) that provides an enumerator
; (that is, for-each-like) interface over some collection. The
; collection can be a file or a database. Iterating over the
; collection may include i/o, and so the possibility of i/o errors. If
; an i/o error occurred, we first make sure we close the files we
; opened to access the collection, and then propagate the exception
; upward.
;
; The library LIB1 provides the following function:
;  for-each-in-coll ITERATEE -> VOID
; where ITERATEE is a procedure ELEM -> VOID.
; The procedure is applied to each element of the collection in turn.
; For this example, we assume that the elements of the collection are
; integers; we simulate file i/o.

(define (for-each-in-coll iteratee)
  (let ((coll-handle
	  (begin (display "opening the collection file") (newline)
	    '(1 2 3 4 5 6 7 #f 9)))) ; we use list to represent
                                     ; the file contents
    (guard
      (exc
	((begin
	   (display "caught exception: closing the collection file")
	   (newline)
	   #f) 'void)			; the exception will be re-raised
	)
      (for-each 
	(lambda (e)
	  (if e (iteratee e) (raise "simulated-io-exc")))
	coll-handle)
      (display "closing the collection file at the end")
      )))

; Here's the example of using for-each-in-coll

(display "An example of using LIB1") (newline)
(guard 
  (exc (else (display "caught: ") (display exc) (newline)))
  (for-each-in-coll (lambda (e) (display e) (newline))))
; opening the collection file
; 1
; 2
; 3
; 4
; 5
; 6
; 7
; caught exception: closing the collection file
; caught: simulated-io-exc


; Suppose we have a library (called LIB2) that provides the cursor-like
; interface to the collection provided by LIB1. The library LIB2 imports
; LIB1 and exports the following procedures:
;
; open-collection UNIT   -> CURSOR or #F
; cursor-element CURSOR  -> VAL
; cursor-next    CURSOR  -> CURSOR or #F
;
; Here #F signifies the end of the collection.
; CURSOR is a data structure (a pair) that contains the current element
; and can be advanced.
; We use shift/reset to invert the enumerator for-each-in-coll to yield
; the cursor. The library LIB2 does no exception handling.

(define (open-collection)
  (reset
    (begin
      (for-each-in-coll (lambda (e) (shift f (cons e f))))
      #f)))
(define (cursor-element cursor) (car cursor))
(define (cursor-next cursor) ((cdr cursor) #f))


; Here's the example of using LIB2

(display "An example of using LIB2") (newline)
(guard 
  (exc (else (display "caught: ") (display exc) (newline)))
  (let ((cursor (open-collection)))
    (let ((cursor			; skip 3 elements
	    (do ((i 0 (+ 1 i)) (cursor cursor (cursor-next cursor)))
	      ((or (not cursor) (>= i 3)) cursor))))
      ; print the rest
      (do ((cursor cursor (cursor-next cursor))) ((not cursor))
	(display (cursor-element cursor))
	(newline)))))


; The main user code
; Suppose the user imports LIB2, opens the cursor, skips a few elements
; and writes the rest in a file. 
; The user handles exceptions to make sure the opened file is closed,
; whether program finishes normally or abnormally.

; The following is a safe file writing procedure: similar to
; call-with-output-file. PROC receives the opened `port'. We open the file,
; execute the PROC, and return its result. The file is closed, no matter
; if PROC returned normally or abnormally.
; For simplicity, we don't actually do any file i/o: me merely display
; messages that we do. Also, the port is a procedure of one argument, 
; the writer.

(define (call-with-file proc)
  (let ((port (begin
		(display "Opening the output file") (newline)
		(lambda (e) 
		  (display "writing to file: ")
		  (display e) (newline)))))
    (guard
      (exc
	((begin
	   (display "caught exception: closing the output file")
	   (newline)
	   #f) 'void)			; the exception will be re-raised
	)
      (proc port)
      (display "closing the output file at the end")
      )))


; As a warm-up, let's see that call-with-file works with LIB1
(display "An example of using LIB1 and call-with-file") (newline)
(guard 
  (exc (else (display "caught: ") (display exc) (newline)))
  (call-with-file
    (lambda (writer)
      (for-each-in-coll (lambda (e) (writer e))))))


; Now we run the master example
(display "An example of using LIB2 and call-with-file") (newline)
(guard 
  (exc (else (display "caught: ") (display exc) (newline)))
  (let ((cursor (open-collection)))
    (let ((cursor			; skip 3 elements
	    (do ((i 0 (+ 1 i)) (cursor cursor (cursor-next cursor)))
	      ((or (not cursor) (>= i 3)) cursor))))
      ; write the rest to the file
      (call-with-file
	(lambda (writer)
	  (do ((cursor cursor (cursor-next cursor))) ((not cursor))
	    (writer (cursor-element cursor))))))))

; The output:
; opening the collection file
; Opening the output file
; writing to file: 4
; writing to file: 5
; writing to file: 6
; writing to file: 7
; caught exception: closing the collection file
; caught: simulated-io-exc

; And here is the problem: the output file remained UNCLOSED!
