; A new implementation of shift/reset
;
; This code requires *non*-winding call/cc -- which is present,
; in particular, in Scheme48. 
;
; This code is Scheme-48 specific. In Scheme48,
; call-with-current-continuation is the R5RS, _winding_ continuation.
; In addition, Scheme 48 has continuation objects (which are not procedures!)
; The function primitive-cwcc captures the current continuation-object,
; and the function with-continuation cont-obj thunk invokes the object.
; The latter function is the equivalent to 'throw' in SML/NJ.
;
; First, do
;	,open signals escapes
;       ,open fluids
;       ,open records threads
;       ,load new-shift.scm
;
; the second line is needed to run the examples at the end of this file.
;
; $Id: new-shift.scm,v 1.2 2006/04/07 07:08:17 oleg Exp $

; we're using primitive-cwcc and with-continuation (aka 'throw')
; forms of Scheme48. The following code is to illustrate how they work

; regular, winding call/cc
(let ((k1 
	(dynamic-wind
	  (lambda () (display "wind1-in") (newline))
	  (lambda () (call-with-current-continuation (lambda (k) k)))
	  (lambda () (display "wind1-out") (newline)))))
  (if (procedure? k1)
    (dynamic-wind
      (lambda () (display "wind2-in") (newline))
      (lambda () (k1 #t))
      (lambda () (display "wind2-out") (newline)))))

; non-winding call/cc
(let ((k1 
	(dynamic-wind
	  (lambda () (display "wind1-in") (newline))
	  (lambda () (primitive-cwcc (lambda (k) k)))
	  (lambda () (display "wind1-out") (newline)))))
  (if (not (boolean? k1))
    (dynamic-wind
      (lambda () (display "wind2-in") (newline))
      (lambda () (with-continuation k1 (lambda () #t)))
      (lambda () (display "wind2-out") (newline)))))


; see the following for details
; scheme48-1.3.tgz!scheme48-1.3/scheme/rts/fluid.scm
; as well as wind.scm in the same directory

(define (get-dynamic-env)
  (record-ref (current-thread) 1))

(define (set-dynamic-env! env)
  (record-set! (current-thread) 1 env))

(define (get-dynamic-point)
  (record-ref (current-thread) 2))

(define (set-dynamic-point! point)
  (record-set! (current-thread) 2 point))

; we should also deal with proposals -- but we skip those for now



(define-syntax reset
  (syntax-rules ()
    ((_ ?e) (*reset (lambda () ?e)))))

(define-syntax shift
  (syntax-rules ()
    ((_ ?k ?e) (*shift (lambda (?k) ?e)))))

; Our meta-continuation is the triple:
;  a function to call with the value (which restores the prev meta-cont)
;  winding continuation of the closest reset
;  and the dynamic env of that reset
(define *meta-continuation*
  (vector
    (lambda (v)
      (error "You forgot the top-level reset...")) ;0
    'no-reset				; 1 full-k
    '()                                 ; 2 dynamic-env
))

; reset the meta-continuation stack
(define (*abort-pure v)
  ((vector-ref *meta-continuation* 0) v))

; execute the thunk in the dynamic env of the closest reset
(define (*abort-env thunk)
  ((vector-ref *meta-continuation* 1) thunk))

(define (*reset thunk)
  (let ((mc *meta-continuation*)
	(env (get-dynamic-env))
	(pt  (get-dynamic-point)))
    (set-dynamic-point! #f)
    (primitive-cwcc
      (lambda (k-pure)
	(*abort-pure
	  ((call-with-current-continuation
	     (lambda (k)
	       (set! *meta-continuation*
		 (vector
		   (lambda (v)
		     (set! *meta-continuation* mc)
		     (set-dynamic-env! env)
		     (set-dynamic-point! pt)
		     (with-continuation k-pure (lambda () v)))
		   k
		   env))
	       thunk))))))))

; compute the list difference between l and lsuffix
(define (unwind-env l lsuffix)
  (if (eq? l lsuffix) '()
    (cons (car l) (unwind-env (cdr l) lsuffix))))
    
(define (*shift f)
  (let* ((parent-env (vector-ref *meta-continuation* 2))
	 (curr-env   (get-dynamic-env))
	 (diff-env   (unwind-env curr-env parent-env)))
    (let ((v (call-with-current-continuation
	       (lambda (k)
		 (*abort-env (lambda ()
			       (f (lambda (v)
					(reset (k v))))))))))
      (set-dynamic-env! (append diff-env (vector-ref *meta-continuation* 2)))
      v)))



;------------------------------------------------------------------------
;			Regular shift tests
(display "regular shift tests") (newline)

(display (+ 10 (reset (+ 2 (shift k (+ 100 (k (k 3))))))))
(newline)
; --> 117

(display (* 10 (reset (* 2 (shift g (reset 
				   (* 5 (shift f (+ (f 1) 1)))))))))
(newline)
; --> 60

(display (let ((f (lambda (x) (shift k (k (k x))))))
	   (+ 1 (reset (+ 10 (f 100))))))
(newline)
; --> 121

(display (reset
	   (let ((x (shift f (cons 'a (f '())))))
	     (shift g x))))
(newline)
; ==> '(a)

(define traverse
  (lambda (xs)
    (letrec ((visit
	       (lambda (xs)
		 (if (null? xs)
		   '()
		   (visit (shift k
			      (cons (car xs) (k (cdr xs)))))))))
      (reset (visit xs)))))

(display "Ex by Oliview Danvy") (newline)
(display (traverse '(1 2 3 4 5)))
(newline)


;------------------------------------------------------------------------
;			New shift and dynvars

(newline) (display "new shift and dynvars") (newline)



(define my-fluid (make-fluid #f))

; de-referencing of my-fluid
(define (dref) (fluid my-fluid))

; dynamic binding: execute thunk in the dynamic env obtained by binding
; my-fluid to new-val
(define (dlet new-val thunk) (let-fluid my-fluid new-val thunk))

; The first test
(define (test1)
  (dlet 1
    (lambda ()
      (list (dref) (dref)
      (dlet 2 
	(lambda ()
	  (list (dref) (dref))))))))

(display (test1)) (newline)
; (1 1 (2 2))

; The counter-example itself

; A sample function to print the current dynamic environment, twice
; (just for the heck of it).
(define (tf title)
  (display title) (newline)
  (display (dref)) (newline)
  (display (dref)) (newline))

; The first test: we capture a delimited continuation that 
; DOES include `dlet 2'
(define (test2)
  (newline) (display "test2") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
	      (reset
		(dlet 2 
		  (lambda ()
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test2)

; the output of the above test:
;
; test2
; inside shift
; 1
; 1
; outside
; 1
; 1
; inside again
; 2
; 2
; outside again
; 1
; 1



; The second test: we now capture a delimited continuation that does NOT
; include `dlet 2'
(define (test3)
  (newline) (display "test3") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
		(dlet 2 
		  (lambda ()
	      (reset
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test3)

; test3
; inside shift
; 2
; 2
; outside
; 1
; 1
; inside again
; 1
; 1
; outside again
; 1
; 1
; 

; Now the results are different -- and just as we expected!

;------------------------------------------------------------------------
; Run the motivating examples from Section 4 of the paper and show
; that in the present implementation, we obtain the expected results.

(newline) (display "Tests from the paper, Section 4")(newline)

(display "Eq4, Section 4.1: ")
(display
  (let ((p (make-fluid #f)))
    (let-fluid p 1 (lambda () (reset (fluid p))))))
(newline)
; ==> 1

(display "Eq5, Section 4.1: ")
(display
  (let ((p (make-fluid #f)))
    (let-fluid p 1
      (lambda () 
	(reset
	  (let-fluid p 2
	    (lambda () (shift f (fluid p)))))))))
(newline)
; ==> 1

(display "Eq6, Section 4.2: ")
(display
  (let ((p (make-fluid #f))
	(q (make-fluid #f)))
    ((lambda (f)
       (let-fluid p 2
	 (lambda ()
	   (let-fluid q 20
	     (lambda () (f 0))))))
     (let-fluid p 1
       (lambda ()
	 (reset
	   (let-fluid q 10
	     (lambda ()
	       ((lambda (x) (+ (fluid p) (fluid q)))
		 (shift f f))))))))))
(newline)
; ==> 12


;------------------------------------------------------------------------
;			New shift and full dynamic-wind actions

(newline) (display "new shift and dynvars") (newline)

(display "test1") (newline)
(dynamic-wind
  (lambda () (display "wind1-in") (newline))
  (lambda () (reset 1))
  (lambda () (display "wind1-out") (newline)))
(newline)

(define (tf1 title) (display title) (newline))

; The analogue of test2 above
(define (testw2)
  (newline) (display "testw2") (newline)
  (dynamic-wind
    (lambda () (display "1-in") (newline))
    (lambda ()
      (let ((f 
	      (reset
		(dynamic-wind
		  (lambda () (display "2-in") (newline))
		  (lambda ()
		    ((shift f (begin (tf1 "inside shift") f)) "inside again"))
		  (lambda () (display "2-out") (newline))
		  ))))
	(tf1 "outside")
	(f tf1)
	(tf1 "outside again")))
    (lambda () (display "1-out") (newline))))
(testw2)

; testw2
; 1-in
; 2-in
; 2-out
; inside shift
; outside
; 2-in
; inside again
; 2-out
; outside again
; 1-out

; The analogue of test3 above
(define (testw3)
  (newline) (display "testw3") (newline)
  (dynamic-wind
    (lambda () (display "1-in") (newline))
    (lambda ()
      (let ((f 
		(dynamic-wind
		  (lambda () (display "2-in") (newline))
		  (lambda ()
	      (reset
		    ((shift f (begin (tf1 "inside shift") f)) "inside again")))
		  (lambda () (display "2-out") (newline))
		  )))
	(tf1 "outside")
	(f tf1)
	(tf1 "outside again")))
    (lambda () (display "1-out") (newline))))
(testw3)

; testw3
; 1-in
; 2-in
; inside shift
; 2-out
; outside
; inside again
; outside again
; 1-out
