; This file illustrates that two, seemingly equivalent implementations
; of shift and reset (with and without trampolining) have observably
; different behavior in the presence of dynamic variables (and, in general,
; dynamic-wind). This code uses (Petite) Chez Scheme native implementation
; of parameter objects.
;
; $Id: trampoline-petite.scm,v 1.2 2006/04/07 01:38:43 oleg Exp $

; Part 1. Common, Danvy/Filinski implementation of shift/reset,
; in terms of call/cc and set!
(load "shift-reset.scm")


(display "shift/reset WITHOUT trampolining") (newline)

; Just to check we have no trampolining
(display "trampoline check") (newline)
(dynamic-wind
  (lambda () (display "in"))
  (lambda () (reset 10))
  (lambda () (display "out")))
(newline)
; prints: inout

(display "reset within dynamic wind") (newline)
(let ((x 0))
  (dynamic-wind
    (lambda () (set! x 1) (display "in"))
    (lambda () (display (reset x)))
    (lambda () (set! x 2) (display "out"))))
(newline)
; prints: in1out

(display "reset within parameterize") (newline)
(let ((p (make-parameter 'top)))
  (parameterize ((p 'inner)) (reset (begin (display (p)) (newline)))))
; prints: inner


(display "(reset ((shift f f))) is not (lambda (f) (f))") (newline)
(let* ((p (make-parameter 1))
       (f (parameterize ((p 2)) (reset ((shift f f))))))
  (display (parameterize ((p 10)) (f p)))
  (newline)
)
; Printed: 2
; although we would have expected 10, because (f p) is just (p)
; and it occurs in the context of (parameterize ((p 10)) ...)


(printf "~nTests from the paper, Section 4~n")
(printf "Eq4, Section 4.1: ~a~n"
  (let ((p (make-parameter 'undefined)))
    (parameterize ((p 1)) (reset (p)))))
; ==> 1

(printf "Eq5, Section 4.1: ~a~n"
  (let ((p (make-parameter 'undefined)))
    (parameterize ((p 1))
      (reset
	(parameterize ((p 2))
	  (shift f (p)))))))
; ==> 2

(printf "Eq6, Section 4.2: ~a~n"
  (let ((p (make-parameter 'undefined))
	(q (make-parameter 'undefined)))
    ((lambda (f)
       (parameterize ((p 2))
	 (parameterize ((q 20))
	   (f 0))))
     (parameterize ((p 1))
      (reset
	(parameterize ((q 10))
	  ((lambda (x) (+ (p) (q)))
	   (shift f f))))))))
; ==> 11



; Part 2. Now add trampolining, which removes the memory leak in the
; implementation of `abort' in Danvy/Filinski code.

; Trampoline, implemented by Chung-chieh Shan
(define with-continuation
  ((call-with-current-continuation
     (lambda (k0)
       (lambda ()
         (lambda (k thunk)
           (k0 (lambda () (k (thunk))))))))))

(define null-continuation (lambda (x) x))

; Trampoline-style, non-leaking abort
(define *abort
  (lambda (thunk)
    (with-continuation null-continuation ;JAR hack
      (lambda ()
        (let ((val (thunk)))
          (*meta-continuation* val))))))

; Now redo all the test
(newline)
(display "shift/reset WITH trampolining") (newline)

; Just to check we have no trampolining
(display "trampoline check") (newline)
(dynamic-wind
  (lambda () (display "in"))
  (lambda () (reset 10))
  (lambda () (display "out")))
(newline)
; prints: inoutinout

(display "reset within dynamic wind") (newline)
(let ((x 0))
  (dynamic-wind
    (lambda () (set! x 1) (display "in"))
    (lambda () (display (reset x)))
    (lambda () (set! x 2) (display "out"))))
(newline)
; prints: inoutin2out

(display "reset within parameterize") (newline)
(let ((p (make-parameter 'top)))
  (parameterize ((p 'inner)) (reset (begin (display (p)) (newline)))))
; prints: top


(display "(reset ((shift f f))) is not (lambda (f) (f))") (newline)
(let* ((p (make-parameter 1))
       (f (parameterize ((p 2)) (reset ((shift f f))))))
  (display (parameterize ((p 10)) (f p)))
  (newline)
)
; Printed: 1
; which is different from the non-trampolined case.
; And still we would have expected 10, because (f p) is just (p)
; and it occurs in the context of (parameterize ((p 10)) ...)

(printf "~nTests from the paper, Section 4~n")
(printf "Eq4, Section 4.1: ~a~n"
  (let ((p (make-parameter 'undefined)))
    (parameterize ((p 1)) (reset (p)))))
; ==> 'undefined

(printf "Eq5, Section 4.1: ~a~n"
  (let ((p (make-parameter 'undefined)))
    (parameterize ((p 1))
      (reset
	(parameterize ((p 2))
	  (shift f (p)))))))
; ==> 'undefined

(printf "Eq6, Section 4.2: ~a~n"
  (let ((p (make-parameter '900))
	(q (make-parameter '9000)))
    ((lambda (f)
       (parameterize ((p 2))
	 (parameterize ((q 20))
	   (f 0))))
     (parameterize ((p 1))
      (reset
	(parameterize ((q 10))
	  ((lambda (x) (+ (p) (q)))
	   (shift f f))))))))
; ==> 910
